/*
	COMPSCI375 Assingment02 (CourseWork 2.2)
	Author: Lily Chia-Chin HSIEH
	UPI: lhsi003
	ID number: 3440845
*/

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.*;
import javax.imageio.*;
import java.util.*;
import java.net.*;

/*
	This class uses the parameter being passed in
	which is the image after threshold alteration
	performs the 4 components algorithm
	colours each component a different colour
	and displays the images
*/
public class DrawFrame extends JFrame {

	private Image imageThresh, imageColoured;	// image objects for image after it has been "threshold", and coloured in, respectively
	private int imageW, imageH;	// width and height of the image, respectively
	private int[] pixels;	// holds the pixels of the image being passed in
	private int[] pixelsAltered;	// holds the pixels for the final outout, ie when it has been given different colours
	private int[][] pixelsOrder;	// groups each same components
	private boolean[][] pixelsVisited;	// boolean 2D array for recording whether a pixels has been visited or not
	private int[][] pixels2D;	// puts the 1D pixel into 2D array for easy referencing
	private PixelGrabber imagepg;
	private int RGBmax = 0xffffffff;	// value for Gmax (object pixels)
	private int count = 1;	// for grouping components
	private Stack<Integer> stackW = new Stack<Integer>();
	private Stack<Integer> stackH = new Stack<Integer>();

	/*
		This method sets the window and calls the necessary methods
		For declarations
	*/
	public DrawFrame(Image imge) {
		setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
		setResizable(false);
        setTitle("Threshold Image(left)<<<<<>>>>>Coloured Image(right)");
		pack();

		imageThresh = imge;
		MediaTracker tracker = new MediaTracker(this);
		tracker.addImage(imageThresh, 0);
		try {
			tracker.waitForID(0);
		} catch (InterruptedException e){}

		imageW = imageThresh.getWidth(this);
		imageH = imageThresh.getHeight(this);

		setBounds(50, 50, imageW*2+50, imageH+50);

		// initialising the array size to the size of the image
		pixels = new int[imageW * imageH];
		pixelsAltered = new int[imageW * imageH];
		pixelsOrder = new int[imageW][imageH];
		pixelsVisited = new boolean[imageW][imageH];
		pixels2D = new int[imageW][imageH];

        imagepg = new PixelGrabber(imageThresh, 0, 0, imageW, imageH, pixels, 0, imageW);
        try {
            imagepg.grabPixels();
        } catch(InterruptedException e) {}

        imagepg = new PixelGrabber(imageThresh, 0, 0, imageW, imageH, pixelsAltered, 0, imageW);
        try {
            imagepg.grabPixels();
        } catch(InterruptedException e) {}

		transTo2D();
		checkComponents();
		colourComponents();
		imageColoured = createImage(new MemoryImageSource(imageW, imageH, pixelsAltered, 0, imageW));

		setVisible(true);
	}

	/*
		Transforming from a 1D array to a 2D array
		for easy referencing later on
	*/
	public void transTo2D() {
		for(int i = 0; i < imageH; i++) {
			for(int k = 0; k < imageW; k++)
				pixels2D[k][i] = pixels[i * imageW + k];
		}
	}

	/*
		Checking whether if each pixel is an object pixels(ie, white)
		and if it has not yet been visited
	*/
	public void checkComponents() {
		for(int i = 0; i < imageH; i ++) {
			for(int k = 0; k < imageW; k++) {
				if((pixels2D[k][i] == RGBmax) && (!pixelsVisited[k][i]))
					groupComponents(i, k);
			}
		}
	}

	/*
		Performing the 4 adjacency component check, in anti-clockwise direction
	*/
	public void groupComponents(int h, int w) {
		pixelsVisited[w][h] = true;
		pixelsOrder[w][h] = count;	// pixel at that point has been added to the current count/component
		stackH.push(h);
		stackW.push(w);
		int temp1 = 0, temp2 = 0;

		// pops h and w one by one
		while(!stackH.empty() && !stackW.empty()) {
			temp2 = (Integer)stackH.pop();
			temp1 = (Integer)stackW.pop();

			// checks the pixel on the right hand side of the current pixel
			// if it is white and has not yet been visited
			// pushes temp1 and temp2 into stackW and stackH respectively
			// sets that point of coordinates to true, ie has been visited
			try{
				if((pixels2D[temp1+1][temp2] == RGBmax) && (!pixelsVisited[temp1+1][temp2])) {
					pixelsOrder[temp1+1][temp2] = count;
					stackH.push(temp2);
					stackW.push(temp1+1);
					pixelsVisited[temp1+1][temp2] = true;
				}
			} catch(Exception ex) {}

			// checks the pixel on top of the current pixel
			// repeats the same processes as above
			try{
				if((pixels2D[temp1][temp2-1] == RGBmax) && (!pixelsVisited[temp1][temp2-1])) {
					pixelsOrder[temp1][temp2-1] = count;
					stackH.push(temp2-1);
					stackW.push(temp1);
					pixelsVisited[temp1][temp2-1] = true;
				}
			} catch(Exception ex) {}

			// checks the pixel on the left hand side of the current pixel
			// repeats the same processes as above
			try{
				if((pixels2D[temp1-1][temp2] == RGBmax) && (!pixelsVisited[temp1-1][temp2])) {
					pixelsOrder[temp1-1][temp2] = count;
					stackH.push(temp2);
					stackW.push(temp1-1);
					pixelsVisited[temp1-1][temp2] = true;
				}
			} catch(Exception ex) {}

			// checks the pixel underneath the current pixel
			// repeats the same processes as above
			try{
				if((pixels2D[temp1][temp2+1] == RGBmax) && (!pixelsVisited[temp1][temp2+1])) {
					pixelsOrder[temp1][temp2+1] = count;
					stackH.push(temp2+1);
					stackW.push(temp1);
					pixelsVisited[temp1][temp2+1] = true;
				}
			} catch(Exception ex) {}
		}
		count++;	// increase the component count by one
	}

	/*
		This method goes through all components
		and colours the same components with the same colour
	*/
	public void colourComponents() {
		Color randColour;
		int newColour;

		for(int i = 0; i < count; i++) {
			// randomly calculates a number between 0 and 255 for each of the R, G, B colours
			// the probability of having 3 randomly generated numbers the same is very very low
			// so assuming that this would not occur
			randColour = new Color((int)(Math.random()*255 + 1), (int)(Math.random()*255 + 1), (int)(Math.random()*255 + 1));
			newColour = randColour.getRGB();

			// nested for loop which checks to see which pixel belongs to the current component
			// and colours that pixel with the current randomly calculated number
			for(int k = 0; k < imageH; k++) {
				for(int j = 0; j < imageW; j++) {
					if(pixelsOrder[j][k] == i && pixels2D[j][k] == RGBmax)
						pixelsAltered[k * imageW + j] = newColour;
				}
			}
		}
	}

	/*
		draws the threshold image and the coloured component image
	*/
	public void paint(Graphics g) {
		g.drawImage(imageThresh, 15, 30, this);
		g.drawImage(imageColoured, imageW+30, 30, this);
	}
}